import scprep
scprep.utils.check_version("keras", "2.2")  # noqa

import keras


def load_mnist():
    (x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()
    x_train = x_train.reshape(x_train.shape[0], -1) / 255
    x_test = x_test.reshape(x_test.shape[0], -1) / 255
    y_train = keras.utils.to_categorical(y_train, num_classes=10)
    y_test = keras.utils.to_categorical(y_test, num_classes=10)
    return x_train, x_test, y_train, y_test


def load_cifar():
    (x_train, y_train), (x_test, y_test) = keras.datasets.cifar10.load_data()
    y_train = keras.utils.to_categorical(y_train, num_classes=10)
    y_test = keras.utils.to_categorical(y_test, num_classes=10)
    x_train = x_train.astype('float32')
    x_test = x_test.astype('float32')
    x_train = x_train.reshape(x_train.shape[0], -1) / 255
    x_test = x_test.reshape(x_test.shape[0], -1) / 255
    return x_train, x_test, y_train, y_test


import graphtools
import graphtools.matrix
from scipy.spatial import distance
from scipy import sparse
import numpy as np
from joblib import Parallel, delayed


def graph_kernel(*args, **kwargs):
    return graphtools.Graph(*args, **kwargs).K


def square_pdist(X):
    return distance.squareform(distance.pdist(X))


def knn_dist(D, interslice_knn, fun=np.median):
    return np.maximum(np.median(np.partition(
        D, interslice_knn, axis=1)[:, interslice_knn]), np.finfo('float').eps)


def distance_to_kernel(D, bandwidth):
    # gaussian kernel
    A = np.exp(-1 * (D / bandwidth)**2)
    A = (A + A.T) / 2
    return A


def _diagonalize_interslice_kernels(interslice_kernels, method='dia'):
    n = interslice_kernels[0].shape[0]
    m = len(interslice_kernels)
    if method == 'dia':
        diags = np.zeros((2*n-1, n*m))
        for vertex, A in enumerate(interslice_kernels):
            diags[(np.tile(np.arange(n, 2*n), n) - np.repeat(np.arange(1, n+1), n), 
                   np.tile(np.arange(0, n*m, m), n) + vertex)] = A.flatten()
        offsets = np.arange(-n*m + m, n*m, m)
        # set main diagonal to 0
        diags[offsets == 0] = 0
        K = sparse.dia_matrix((diags, offsets), shape=(n*m, n*m))
    else:
        assert method == 'csr'
        # this is inefficient
        K = sparse.csr_matrix((n*m, n*m))
        for vertex, A in enumerate(interslice_kernels):
            K = graphtools.matrix.set_submatrix(
                K, np.arange(n) * m + vertex, np.arange(n) * m + vertex, A)
        # set main diagonal to 0
        K = graphtools.matrix.set_diagonal(K, 0)
    return K
        


def _multislice_kernel(data,
                       intraslice_knn=2,
                       interslice_knn=25,
                       decay=5,
                       n_pca=100,
                       p=list,
                       kernel_fn=graph_kernel,
                       pdist_fn=square_pdist,
                       knn_fn=knn_dist,
                       kernel_dist_fn=distance_to_kernel,
                       **kwargs):
    n = data.shape[0]
    m = data.shape[1]
    N = n * m
    K = sparse.lil_matrix((N, N))

    kernels = p(kernel_fn(
        x, knn=intraslice_knn, decay=decay, **kwargs, n_jobs=1)
        for x in data)
    for i, G_K in enumerate(kernels):
        # plug into K
        K = graphtools.matrix.set_submatrix(K, np.arange(
            i * m, (i + 1) * m), np.arange(i * m, (i + 1) * m), G_K)

    # set interslice fixed bandwidth
    interslice_dist = p(pdist_fn(data[:, vertex, :])
                        for vertex in range(data.shape[1]))
    bandwidths = p(knn_fn(D, interslice_knn)
                   for D in interslice_dist)
    bandwidth = np.mean(bandwidths)
    interslice_kernels = p(kernel_dist_fn(
        D, bandwidth)
        for D in interslice_dist)

    K = K.tocsr() + _diagonalize_interslice_kernels(interslice_kernels).tocsr()
        
    return K


def multislice_kernel(data,
                      intraslice_knn=2,
                      interslice_knn=25,
                      decay=5,
                      n_pca=100,
                      n_jobs=20, **kwargs):
    n = data.shape[0]
    m = data.shape[1]

    if n_pca is not None and n_pca < data.shape[2]:
        data = data.reshape(n * m, -1)
        data = graphtools.base.Data(data, n_pca=n_pca).data_nu
        data = data.reshape(n, m, n_pca)

    # build within slice graphs
    if n_jobs == 1:
        K = _multislice_kernel(data,
                               intraslice_knn=intraslice_knn,
                               interslice_knn=interslice_knn,
                               decay=decay,
                               n_pca=n_pca)
    else:
        with Parallel(n_jobs=n_jobs) as p:
            K = _multislice_kernel(
                data,
                intraslice_knn=intraslice_knn,
                interslice_knn=interslice_knn,
                decay=decay,
                n_pca=n_pca,
                p=p,
                kernel_fn=delayed(graph_kernel),
                pdist_fn=delayed(square_pdist),
                knn_fn=delayed(knn_dist),
                kernel_dist_fn=delayed(distance_to_kernel))
    return K.tocsr()


def DM(G, t=1, n_components=2):
    # symmetric affinity matrix
    diff_aff = sparse.csr_matrix(G.diff_aff)
    # symmetrize to remove numerical error
    diff_aff = (diff_aff + diff_aff.T) / 2
    # svd
    U, S, _ = sparse.linalg.svds(diff_aff, k=n_components + 1)
    # order in terms of smallest eigenvalue
    U, S = U[:, ::-1], S[::-1]
    # get first eigenvector
    u1 = U[:, 0][:, None]
    # ensure non-zero
    zero_idx = np.abs(u1) <= np.finfo(float).eps
    u1[zero_idx] = (np.sign(u1[zero_idx]) * np.finfo(
        float).eps).reshape(-1)
    # normalize by first eigenvector
    U = U / u1
    # renormalize to length 1
    U = U / np.linalg.norm(U, axis=0)
    # drop first eigenvector
    U, S = U[:, 1:], S[1:]
    # power eigenvalues
    S = np.power(S, t)
    # weight U by eigenvalues
    dm = U.dot(np.diagflat(S))
    return dm



import numpy as np
import m_phate
import graphtools
import warnings
from scipy import sparse
from parameterized import parameterized

def test_diagonalize_interslice_kernels():
    n = 15
    m = 8
    kernels = [np.arange(n**2).reshape(n,n) + i*n**2 for i in range(m)]
    
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=sparse.SparseEfficiencyWarning)
        K = m_phate.kernel._diagonalize_interslice_kernels(kernels, method='csr')
    D = m_phate.kernel._diagonalize_interslice_kernels(kernels, method='dia')
    assert (D.tocsr() - K).nnz == 0

@parameterized(
    [(1,), (-1,)])
def test_m_phate(n_jobs):
    # create fake data
    n_time_steps = 50
    n_points = 20
    n_dim = 10
    n_pca = 5
    np.random.seed(42)
    data = np.cumsum(np.random.normal(
        0, 1, (n_time_steps, n_points, n_dim)), axis=0)

    # embedding
    m_phate_op = m_phate.M_PHATE(n_jobs=n_jobs, verbose=0, n_pca=n_pca)
    m_phate_data = m_phate_op.fit_transform(data)

    assert m_phate_data.shape[0] == n_points * n_time_steps
    assert m_phate_data.shape[1] == 2

    m_phate_op.set_params(intraslice_knn=m_phate_op.intraslice_knn)
    assert isinstance(m_phate_op.graph, graphtools.base.BaseGraph)
    m_phate_op.set_params(interslice_knn=m_phate_op.interslice_knn)
    assert isinstance(m_phate_op.graph, graphtools.base.BaseGraph)
    m_phate_op.set_params(n_svd=m_phate_op.n_svd)
    assert isinstance(m_phate_op.graph, graphtools.base.BaseGraph)

    G = m_phate_op.graph
    m_phate_op.set_params(intraslice_knn=m_phate_op.intraslice_knn+1)
    assert m_phate_op.graph is None
    m_phate_op.graph = G
    m_phate_op.set_params(interslice_knn=m_phate_op.interslice_knn+1)
    assert m_phate_op.graph is None

@parameterized(
    [(2,), (3,)])
def test_multislice_kernel(intraslice_knn):
    # create fake data
    n_time_steps = 50
    n_points = 20
    n_dim = 10
    np.random.seed(42)
    data = np.cumsum(np.random.normal(
        0, 1, (n_time_steps, n_points, n_dim)), axis=0)
    kernel = m_phate.kernel.multislice_kernel(m_phate.utils.normalize(data), 
                                              intraslice_knn=intraslice_knn,
                                              decay=None)

    nnz = 0
    # intraslice kernel
    for t in range(n_time_steps):
        subkernel = kernel[t*n_points:(t+1)*n_points][:,t*n_points:(t+1)*n_points]
        assert subkernel.sum() == n_points * (intraslice_knn+1)
        nnz += subkernel.nnz

    # interslice kernel
    for i in range(n_points):
        subkernel = kernel[i::n_points][:,i::n_points]
        assert subkernel.nnz == n_time_steps ** 2
        nnz += subkernel.nnz

    # diagonal is double counted
    nnz -= kernel.shape[0]
    # everything else should be zero
    assert nnz == kernel.nnz

    # check this passes through phate op
    m_phate_op = m_phate.M_PHATE(intraslice_knn=intraslice_knn,
                                 decay=None, verbose=0)
    m_phate_data = m_phate_op.fit_transform(data)

    # threshold
    kernel.data[kernel.data < 1e-4] = 0

    assert m_phate_data.shape[0] == n_points * n_time_steps
    assert m_phate_data.shape[1] == 2
    np.testing.assert_allclose((m_phate_op.graph.kernel - kernel).data, 0,
                               rtol=0, atol=1e-14)


def test_dm():
    # create fake data
    n_time_steps = 50
    n_points = 20
    n_dim = 10
    np.random.seed(42)
    data = np.cumsum(np.random.normal(
        0, 1, (n_time_steps, n_points, n_dim)), axis=0)
    kernel = m_phate.kernel.multislice_kernel(m_phate.utils.normalize(data))
    dm = m_phate.kernel.DM(graphtools.Graph(kernel, precomputed='affinity'))
    assert dm.shape == (n_time_steps * n_points, 2)


def test_normalize():
    # create fake data
    n_time_steps = 50
    n_points = 20
    n_dim = 10
    np.random.seed(42)
    data = np.cumsum(np.random.normal(
        0, 1, (n_time_steps, n_points, n_dim)), axis=0)
    data_norm = m_phate.utils.normalize(data)
    np.testing.assert_allclose(data_norm.mean(axis=2), 0, rtol=0, atol=1e-15)
    np.testing.assert_allclose(data_norm.std(axis=2), 1, rtol=0, atol=1e-15)



import m_phate.train
import os
import numpy as np
import keras
from parameterized import parameterized

def test_config():
    config = m_phate.train.build_config(limit_gpu_fraction=0)
    assert config.device_count['GPU'] == 0
    config = m_phate.train.build_config(limit_gpu_fraction=0.5)
    assert config.gpu_options.per_process_gpu_memory_fraction == 0.5
    assert config.gpu_options.allow_growth
    config = m_phate.train.build_config(limit_cpu_fraction=2)
    assert config.intra_op_parallelism_threads == 2
    assert config.inter_op_parallelism_threads == 2
    config = m_phate.train.build_config(limit_cpu_fraction=0.5)
    assert config.intra_op_parallelism_threads == os.cpu_count() // 2
    assert config.inter_op_parallelism_threads == os.cpu_count() // 2
    config = m_phate.train.build_config(limit_cpu_fraction=-2)
    assert config.intra_op_parallelism_threads == os.cpu_count() - 1, (config.intra_op_parallelism_threads, os.cpu_count() - 1)
    assert config.inter_op_parallelism_threads == os.cpu_count() - 1
    config = m_phate.train.build_config(limit_cpu_fraction=0)
    assert config.intra_op_parallelism_threads == 1
    assert config.inter_op_parallelism_threads == 1

def test_trace_1_layer():
    # create data
    n_points = 100
    n_batch = 10
    n_trace = 10
    n_dim = 50
    n_hidden = 20
    x = np.random.normal(0, 1, [n_points, n_dim])

    # select trace examples
    trace_idx = np.arange(n_trace)
    trace_data = x[trace_idx]

    # build neural network
    inputs = keras.layers.Input(
        shape=(n_dim,), dtype='float32', name='inputs')
    h1 = keras.layers.Dense(n_hidden, name='h1')(inputs)
    outputs = keras.layers.Dense(n_dim, name='output_all')(h1)

    # build trace model helper
    model_trace = keras.models.Model(inputs=inputs, outputs=[h1])
    trace = m_phate.train.TraceHistory(trace_data, model_trace)
    batch_trace = m_phate.train.BatchTraceHistory(trace_data, model_trace)

    # compile network
    model = keras.models.Model(inputs=inputs, outputs=outputs)
    model.compile(optimizer='adam', loss='mean_squared_error')

    # train network
    model.fit(x, x, batch_size=n_points // n_batch, epochs=1,
              verbose=0, callbacks=[trace, batch_trace])

    assert len(trace.trace) == 1
    assert trace.trace[0].shape == (n_hidden, n_trace), trace.trace[0].shape
    assert len(batch_trace.trace) == n_batch, len(batch_trace.trace)
    assert np.all([t.shape == (n_hidden, n_trace) for t in batch_trace.trace])

def test_trace_2_layer():
    # create data
    n_points = 100
    n_batch = 10
    n_trace = 10
    n_dim = 50
    n_hidden = 20
    x = np.random.normal(0, 1, [n_points, n_dim])

    # select trace examples
    trace_idx = np.arange(n_trace)
    trace_data = x[trace_idx]

    # build neural network
    inputs = keras.layers.Input(
        shape=(n_dim,), dtype='float32', name='inputs')
    h1 = keras.layers.Dense(n_hidden, name='h1')(inputs)
    h2 = keras.layers.Dense(n_hidden * 2, name='h2')(h1)
    outputs = keras.layers.Dense(n_dim, name='output_all')(h2)

    # build trace model helper
    model_trace = keras.models.Model(inputs=inputs, outputs=[h1, h2])
    trace = m_phate.train.TraceHistory(trace_data, model_trace)
    batch_trace = m_phate.train.BatchTraceHistory(trace_data, model_trace)

    # compile network
    model = keras.models.Model(inputs=inputs, outputs=outputs)
    model.compile(optimizer='adam', loss='mean_squared_error')

    # train network
    model.fit(x, x, batch_size=n_points // n_batch, epochs=1,
              verbose=0, callbacks=[trace, batch_trace])

    assert len(trace.trace) == 1, len(trace.trace)
    assert trace.trace[0].shape == (3*n_hidden, n_trace), trace.trace[0].shape
    assert len(batch_trace.trace) == n_batch, len(batch_trace.trace)
    assert np.all([t.shape == (3*n_hidden, n_trace) for t in batch_trace.trace])


import phate
import graphtools
import tasklogger
import numpy as np

from . import kernel, utils

_logger = tasklogger.get_tasklogger("graphtools")


class M_PHATE(phate.PHATE):
    """Multislice PHATE operator which performs dimensionality reduction.

    M-PHATE embeds tensors of changing data in two or three dimensions
    for the visualization of the evolution of a dynamical system over time
    as described in Gigante et al, 2019 [1]_ for the visualization of
    the evolution of neural networks throughout learning.

    Parameters
    ----------

    n_components : int, optional, default: 2
        number of dimensions in which the data will be embedded

    intraslice_knn : int, optional, default: 2
        number of nearest neighbors on which to build intraslice kernels

    interslice_knn : int, optional, default: 25
        number of nearest neighbors on which to build interslice kernels

    decay : int, optional, default: 5
        sets decay rate of kernel tails.
        If None, alpha decaying kernel is not used

    n_landmark : int, optional, default: 4000
        number of landmarks to use in fast PHATE

    t : int, optional, default: 'auto'
        power to which the diffusion operator is powered.
        This sets the level of diffusion. If 'auto', t is selected
        according to the knee point in the Von Neumann Entropy of
        the diffusion operator

    gamma : float, optional, default: 0
        Informational distance constant between -1 and 1.
        `gamma=1` gives the PHATE log potential, `gamma=0` gives
        a square root potential.

    normalize : bool, optional, default: True
        If True, z-score the data

    mds_solver : {'smacof', 'sgd'}, optional (default: 'smacof')
        which solver to use for metric MDS. SGD is substantially faster,
        but produces slightly less optimal results.

    n_pca : int, optional, default: 100
        Number of principal components to use for calculating
        neighborhoods. For extremely large datasets, using
        n_pca < 20 allows neighborhoods to be calculated in
        roughly log(n_samples) time.

    n_svd : int, optional, default: 100
        Number of singular vectors to use for calculating
        landmarks.

    n_jobs : integer, optional, default: 1
        The number of jobs to use for the computation.
        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging.
        For n_jobs below -1, (n_cpus + 1 + n_jobs) are used. Thus for
        n_jobs = -2, all CPUs but one are used

    random_state : integer or numpy.RandomState, optional, default: None
        The generator used to initialize SMACOF (metric, nonmetric) MDS
        If an integer is given, it fixes the seed
        Defaults to the global `numpy` random number generator

    verbose : `int` or `boolean`, optional (default: 1)
        If `True` or `> 0`, print status messages

    phate_kwargs : additional keyword arguments for phate.PHATE

    Attributes
    ----------

    X : array-like, shape=[n_samples, n_dimensions]

    embedding : array-like, shape=[n_samples, n_components]
        Stores the position of the dataset in the embedding space

    diff_op :  array-like, shape=[n_samples, n_samples] or [n_landmark, n_landmark]
        The diffusion operator built from the graph

    graph : graphtools.base.BaseGraph
        The graph built on the input data

    optimal_t : int
        The automatically selected t, when t = 'auto'.
        When t is given, optimal_t is None.

    Examples
    --------
    >>> import m_phate
    >>> m_phate_operator = m_phate.M_PHATE()
    >>> phate.plot.scatter2d(tree_phate, c=tree_clusters)

    References
    ----------
    .. [1] Gigante S, Charles AS, Krishnaswamy S and Mishne G (2019),
        *Visualizing the PHATE of Neural Networks*,
        `arXiv <http://arxiv.org/abs/>`_.
    """

    def __init__(self, n_components=2,
                 intraslice_knn=2, interslice_knn=25,
                 decay=5, t='auto', gamma=0, n_landmark=4000,
                 normalize=True, mds_solver='smacof', n_pca=100, n_svd=100,
                 n_jobs=-2, random_state=None, verbose=1,
                 knn=None,
                 **phate_kwargs):
        if knn is not None:
            warnings.warn("Argument `knn` is ambiguous and ignored. "
                          "Use `intraslice_knn` or `interslice_knn`.",
                          UserWarning)
        self.interslice_knn = interslice_knn
        self.n_svd = n_svd
        self.normalize = normalize
        return super().__init__(
            n_components=n_components,
            knn=intraslice_knn,
            mds_solver=mds_solver,
            decay=decay, t=t,
            n_pca=n_pca, gamma=gamma,
            n_landmark=n_landmark,
            n_jobs=n_jobs, random_state=random_state,
            verbose=verbose, **phate_kwargs)

    @property
    def intraslice_knn(self):
        return self.knn

    def fit(self, X):
        if not len(X.shape) == 3:
            raise ValueError("Expected X to be a tensor with three dimensions."
                             " Got shape {}".format(X.shape))

        if self.normalize:
            X = utils.normalize(X)

        with _logger.task("multislice kernel"):
            K = kernel.multislice_kernel(X,
                                         intraslice_knn=self.intraslice_knn,
                                         interslice_knn=self.interslice_knn,
                                         decay=self.decay,
                                         n_pca=self.n_pca,
                                         distance=self.knn_dist,
                                         n_jobs=self.n_jobs)
        with _logger.task("graph and diffusion operator"):
            n_landmark = self.n_landmark if self.n_landmark is not None and self.n_landmark < K.shape[0] else None
            self.graph = graphtools.Graph(
                K,
                precomputed="affinity",
                n_landmark=n_landmark,
                n_svd=self.n_svd,
                n_jobs=self.n_jobs,
                verbose=self.verbose,
                random_state=self.random_state,
                **(self.kwargs))
            self.diff_op
        result = super().fit(self.graph)
        return result

    def fit_transform(self, X, **kwargs):
        """Computes the diffusion operator and the position of the cells in the
        embedding space

        Parameters
        ----------
        X : array, shape=[n_samples, n_features]
            input data with `n_samples` samples and `n_dimensions`
            dimensions. Accepted data types: `numpy.ndarray`,
            `scipy.sparse.spmatrix`, `pd.DataFrame`

        kwargs : further arguments for `M_PHATE.transform()`
            Keyword arguments as specified in :func:`~m_phate.M_PHATE.transform`

        Returns
        -------
        embedding : array, shape=[n_samples, n_dimensions]
            The cells embedded in a lower dimensional space using PHATE
        """
        with _logger.task('M-PHATE'):
            self.fit(X)
            embedding = self.transform(**kwargs)
        return embedding

    def _check_params(self):
        """Check M-PHATE parameters

        This allows us to fail early - otherwise certain unacceptable
        parameter choices, such as mds='mmds', would only fail after
        minutes of runtime.

        Raises
        ------
        ValueError : unacceptable choice of parameters
        """
        phate.utils.check_int(interslice_knn=self.interslice_knn,
                              n_svd=self.n_svd)
        phate.utils.check_positive(interslice_knn=self.interslice_knn,
                                   n_svd=self.n_svd)
        return super()._check_params()

    def set_params(self, **params):
        """Set the parameters on this estimator.

        Any parameters not given as named arguments will be left at their
        current value.

        Parameters
        ----------

        n_components : int, optional, default: 2
            number of dimensions in which the data will be embedded

        intraslice_knn : int, optional, default: 2
            number of nearest neighbors on which to build intraslice kernels

        interslice_knn : int, optional, default: 25
            number of nearest neighbors on which to build interslice kernels

        decay : int, optional, default: 5
            sets decay rate of kernel tails.
            If None, alpha decaying kernel is not used

        n_landmark : int, optional, default: 4000
            number of landmarks to use in fast PHATE

        t : int, optional, default: 'auto'
            power to which the diffusion operator is powered.
            This sets the level of diffusion. If 'auto', t is selected
            according to the knee point in the Von Neumann Entropy of
            the diffusion operator

        gamma : float, optional, default: 0
            Informational distance constant between -1 and 1.
            `gamma=1` gives the PHATE log potential, `gamma=0` gives
            a square root potential.

        n_pca : int, optional, default: 100
            Number of principal components to use for calculating
            neighborhoods. For extremely large datasets, using
            n_pca < 20 allows neighborhoods to be calculated in
            roughly log(n_samples) time.

        n_svd : int, optional, default: 100
            Number of singular vectors to use for calculating
            landmarks.

        n_jobs : integer, optional, default: 1
            The number of jobs to use for the computation.
            If -1 all CPUs are used. If 1 is given, no parallel computing code is
            used at all, which is useful for debugging.
            For n_jobs below -1, (n_cpus + 1 + n_jobs) are used. Thus for
            n_jobs = -2, all CPUs but one are used

        random_state : integer or numpy.RandomState, optional, default: None
            The generator used to initialize SMACOF (metric, nonmetric) MDS
            If an integer is given, it fixes the seed
            Defaults to the global `numpy` random number generator

        verbose : `int` or `boolean`, optional (default: 1)
            If `True` or `> 0`, print status messages
        """
        reset_kernel = False

        if 'interslice_knn' in params and params['interslice_knn'] != self.interslice_knn:
            self.interslice_knn = params['interslice_knn']
            reset_kernel = True
            del params['interslice_knn']

        if 'intraslice_knn' in params:
            params['knn'] = params['intraslice_knn']
            del params['intraslice_knn']

        if 'n_svd' in params and params['n_svd'] != self.n_svd:
            self._set_graph_params(n_svd=params['n_svd'])
            self.n_svd = params['n_svd']
            del params['n_svd']

        if reset_kernel:
            # can't reset the graph kernel without making a new graph
            self._reset_graph()

        return super().set_params(**params)



import numpy as np
import os
import scprep

scprep.utils.check_version("keras", "2.2")  # noqa
scprep.utils.check_version("tensorflow", "1.13")  # noqa

import keras

try:
    from tensorflow import GPUOptions, ConfigProto
except ImportError:
    from tensorflow.compat.v1 import GPUOptions, ConfigProto

def build_config(limit_gpu_fraction=0.2, limit_cpu_fraction=10):
    if limit_gpu_fraction > 0:
        os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        gpu_options = GPUOptions(
            allow_growth=True,
            per_process_gpu_memory_fraction=limit_gpu_fraction)
        config = ConfigProto(gpu_options=gpu_options)
    else:
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
        config = ConfigProto(device_count={'GPU': 0})
    if limit_cpu_fraction is not None:
        if limit_cpu_fraction == 0:
            cpu_count = 1
        if limit_cpu_fraction < 0:
            # -2 gives all CPUs except 1
            cpu_count = max(
                1, int(os.cpu_count() + limit_cpu_fraction + 1))
        elif limit_cpu_fraction < 1:
            # 0.5 gives 50% of available CPUs
            cpu_count = max(
                1, int(os.cpu_count() * limit_cpu_fraction))
        else:
            # 2 gives 2 CPUs
            cpu_count = int(limit_cpu_fraction)
        config.inter_op_parallelism_threads = cpu_count
        config.intra_op_parallelism_threads = cpu_count
        os.environ['OMP_NUM_THREADS'] = str(1)
        os.environ['MKL_NUM_THREADS'] = str(cpu_count)
    return config


class _History(keras.callbacks.Callback):

    def __init__(self, data, model, save_weights=False, *args, **kwargs):
        self.trace_data = data
        self.trace_model = model
        self.save_weights = save_weights
        if save_weights:
            self.weights = []
        self.trace = []
        super().__init__(*args, **kwargs)

    def _record_trace(self):
        if len(self.trace_model.outputs) > 1:
            self.trace.append(
                np.vstack([data.T for data in self.trace_model.predict(
                    self.trace_data)]))
        else:
            self.trace.append(self.trace_model.predict(
                    self.trace_data).T)
        if self.save_weights:
            self.weights.append(self.trace_model.layers[1].get_weights()[0])


class TraceHistory(_History):

    def on_epoch_end(self, epoch, logs):
        self._record_trace()
        return super().on_epoch_end(epoch, logs)


class BatchTraceHistory(_History):

    def on_batch_end(self, epoch, logs):
        self._record_trace()
        return super().on_batch_end(epoch, logs)



import os
import sys
from setuptools import setup, find_packages

install_requires = [
    'numpy>=1.16.0',
    'scipy>=1.1.0',
    'scikit-learn>=0.19.1',
    'pandas>=0.25',
    'joblib',
    'scprep>=1.0',
    'phate>=1.0.0',
    'graphtools>=1.5',
    "tasklogger>=1.0",
]

test_requires = [
    'nose2',
    'coverage',
    'coveralls',
    'keras>=2.2',
    'tensorflow>=1.13,<2.0',
    'parameterized',
]

doc_requires = [
    'sphinx<=1.8.5',
    'sphinxcontrib-napoleon',
    'autodocsumm',
    'ipykernel',
    'nbsphinx',
]

if sys.version_info[:2] < (3, 5):
    raise RuntimeError("Python version >=3.5 required.")
elif sys.version_info[:2] < (3, 6):
    install_requires += ['matplotlib>=3.0,<3.1']
else:
    install_requires += ['matplotlib>=3.0']

version_py = os.path.join(os.path.dirname(
    __file__), 'm_phate', 'version.py')
version = open(version_py).read().strip().split(
    '=')[-1].replace('"', '').strip()

readme = open('README.md').read()

setup(name='m_phate',
      version=version,
      description='m-phate',
      author='Scott Gigante, Yale University',
      author_email='scott.gigante@yale.edu',
      packages=find_packages(),
      license='GNU General Public License Version 3',
      install_requires=install_requires,
      extras_require={'test': test_requires,
                      'doc': doc_requires},
      test_suite='nose2.collector.collector',
      long_description=readme,
      long_description_content_type='text/markdown',
      url='https://github.com/scottgigante/m-phate',
      download_url="https://github.com/scottgigante/m-phate/archive/v{}.tar.gz".format(
          version),
      keywords=['big-data',
                'computational-biology',
                ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Framework :: Jupyter',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Scientific/Engineering :: Visualization',
      ]
      )



import m_phate.data
from parameterized import parameterized

@parameterized(
    [(m_phate.data.load_mnist,), 
     (m_phate.data.load_cifar,)])
def test_data(data_fn):
    x_train, x_test, y_train, y_test = data_fn()
    assert x_train.shape[0] == y_train.shape[0]
    assert x_test.shape[0] == y_test.shape[0]
    assert x_train.shape[1] == x_test.shape[1]
    assert y_train.shape[1] == y_test.shape[1]
    assert x_train.shape[0] > x_test.shape[0]
